Package: ao
Title: Alternating Optimization
Version: 0.3.3
Authors@R: 
    person(given = "Lennart",
           family = "Oelschl\U00E4ger",
           role = c("aut", "cre"),
           email = "oelschlaeger.lennart@gmail.com",
           comment = c(ORCID = "0000-0001-5421-9313"))
Description: Alternating optimization of (high-dimensional) functions is an 
    iterative procedure for optimizing jointly over all parameters by 
    alternately optimizing parameter subsets.
URL: https://loelschlaeger.de/ao/, https://github.com/loelschlaeger/ao/
BugReports: https://github.com/loelschlaeger/ao/issues
License: GPL-3
Encoding: UTF-8
RoxygenNote: 7.3.1
Imports: checkmate, cli, oeli, utils
Suggests: knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
Depends: R (>= 4.0.0), optimizeR (>= 1.0.4)
NeedsCompilation: no
Packaged: 2024-02-29 20:05:27 UTC; Oelschlaeger
Author: Lennart Oelschläger [aut, cre]
    (<https://orcid.org/0000-0001-5421-9313>)
Maintainer: Lennart Oelschläger <oelschlaeger.lennart@gmail.com>
Repository: CRAN
Date/Publication: 2024-02-29 21:12:46 UTC
Built: R 4.2.3; ; 2024-03-04 12:19:35 UTC; unix
