# Generated by roxygen2: do not edit by hand

export(ao)
importFrom(R6,R6Class)
importFrom(checkmate,assert_int)
importFrom(checkmate,assert_number)
importFrom(checkmate,test_class)
importFrom(checkmate,test_flag)
importFrom(checkmate,test_function)
importFrom(checkmate,test_list)
importFrom(checkmate,test_number)
importFrom(cli,cat_line)
importFrom(cli,cli_abort)
importFrom(cli,style_hyperlink)
importFrom(future.apply,future_lapply)
importFrom(oeli,match_arg)
importFrom(oeli,test_numeric_vector)
importFrom(optimizeR,Objective)
importFrom(optimizeR,Optimizer)
importFrom(progressr,progressor)
importFrom(stats,rnorm)
importFrom(utils,packageVersion)
